import hydra
import wandb
import random
import os
import numpy as np
import gymnasium as gym
from pathlib import Path
from datetime import datetime
from omegaconf import DictConfig

import torch
import torch.distributions as td
import torch.nn.functional as F
from torch.utils.data import DataLoader

from model import DecisionRvs, DecisionV
from utils import MaxEpisodicDataset, KMeansEpisodicDataset, VisionKMeansEpisodicDataset, VisionMaxEpisodicDataset, get_test_start_state_goals, get_lr, AntmazeWrapper, PreprocessObservationWrapper 
from gymnasium.wrappers import PixelObservationWrapper
from torch import distributions, nn
from torch.nn import functional as F
from functools import partial
from typing import List, Optional, Type, Any, Dict, Optional, Tuple, Union

def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
         
def eval_env(cfg, model, v_model, device, render=False):
    if render:
        render_mode = 'human'
    else:
        render_mode = None

    if cfg.env_name in ['PointMaze_UMaze-v3', 'PointMaze_Medium-v3', 'PointMaze_Large-v3']:
        if cfg.vision:
            env = gym.make(cfg.env_name, continuing_task=False, render_mode='rgb_array')
            DEFAULT_CAMERA_CONFIG = {
            "distance": 14 if len(env.maze.maze_map) > 8 else 8.8,
            "elevation": -90,
            "lookat": [0, 0, 0,],
            }
            env.point_env.mujoco_renderer.default_cam_config = DEFAULT_CAMERA_CONFIG
            env = PixelObservationWrapper(env, pixels_only=False)
            env = PreprocessObservationWrapper(env, shape=64, grayscale=True)
        else:
            env = env = gym.make(cfg.env_name, continuing_task=False, render_mode=render_mode)
    elif cfg.env_name in ['AntMaze_UMaze-v4', 'AntMaze_Medium-v4', 'AntMaze_Large-v4']:
        env = AntmazeWrapper(env = gym.make(cfg.env_name, continuing_task=False, render_mode=render_mode))
    else:
        raise NotImplementedError    

    test_start_state_goal = get_test_start_state_goals(cfg)
    
    model.eval()
    v_model.eval()
    
    results = dict()
    with torch.no_grad():
        mega_reward = 0
        for ss_g in test_start_state_goal:
            total_reward = 0
            total_timesteps = 0
            
            print(ss_g['name'] + ':')
            for _ in range(cfg.num_eval_ep):
                obs, _ = env.reset(options=ss_g)

                running_state = torch.tensor(obs['observation'], dtype=torch.float32, device=device).view(1, -1) 
                #print("running_state.shape:",running_state.shape)
                # running_proprio = torch.tensor(obs['observation'][2:], dtype=torch.float32, device=device).view(1, -1)
                target_goal = torch.tensor(obs['desired_goal'], dtype=torch.float32, device=device).view(1, -1)
                #print("target_goal.shape:",target_goal.shape)
                #act = torch.zeros((model.act_dim,), dtype=torch.float32, device=device).view(1, -1)
                #print("act.shape:",act.shape)

                done = False

                for t in range(env.spec.max_episode_steps):
                    total_timesteps += 1
                
                    v = v_model.forward(
                            running_state,
                            # running_proprio,
                            #act,
                            target_goal,
                            )

                    act_preds = model.forward(
                            running_state,
                            # running_proprio,
                            target_goal,
                            v,
                            )
                    act = act_preds[0].detach()

                    obs, running_reward, done, _, _ = env.step(act.cpu().numpy())

                    running_state = torch.tensor(obs['observation'], dtype=torch.float32, device=device).view(1, -1)
                    # running_proprio = torch.tensor(obs['observation'][2:], dtype=torch.float32, device=device).view(1, -1)
                    target_goal = torch.tensor(obs['desired_goal'], dtype=torch.float32, device=device).view(1, -1)
                    act = act.view(1, -1)
                    
                    total_reward += running_reward

                    if done:
                        break
                
                print('Achievied goal: ', tuple(obs['achieved_goal'].tolist()))
                print('Desired goal: ', tuple(obs['desired_goal'].tolist()))
            
            print("=" * 60)
            mega_reward += total_reward
            results['eval/' + str(ss_g['name']) + '_avg_reward'] = total_reward / cfg.num_eval_ep
            results['eval/' + str(ss_g['name']) + '_avg_ep_len'] = total_timesteps / cfg.num_eval_ep
        
        results['eval/avg_reward'] = mega_reward / (cfg.num_eval_ep * len(test_start_state_goal))
        env.close()
    return results

class VAE(nn.Module):
    # Vanilla Variational Auto-Encoder
    def __init__(self,state_dim: int,
                    action_dim: int,
                    goal_dim: int,
                    latent_dim: int,
                    max_action: float,
                    hidden_dim: int = 750):

        super(VAE, self).__init__()
        if latent_dim is None:
            latent_dim = 2 * action_dim
        #
        self.encoder_shared = nn.Sequential( nn.Linear(state_dim + action_dim + goal_dim, hidden_dim),
                                             nn.ReLU(),
                                             nn.Linear(hidden_dim, hidden_dim),
                                             nn.ReLU() )

        self.mean    = nn.Linear(hidden_dim, latent_dim)
        self.log_std = nn.Linear(hidden_dim, latent_dim)
        self.decoder = nn.Sequential( nn.Linear(state_dim + action_dim + latent_dim, hidden_dim),
                                      nn.ReLU(),
                                      nn.Linear(hidden_dim, hidden_dim),
                                      nn.ReLU(),
                                      nn.Linear(hidden_dim, goal_dim),
                                      nn.Sigmoid())
        
        self.max_action = max_action
        self.latent_dim = latent_dim
        self.device =  torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def forward(self, state : torch.Tensor,
                      action: torch.Tensor,
                      goal  : torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        mean, std = self.encode(state, action, goal)
        z = mean + std * torch.randn_like(std)
        u = self.decode(state, action, z)
        return u, mean, std

    def importance_sampling_estimator(self, state: torch.Tensor,
                                            action: torch.Tensor,
                                            goal:torch.Tensor,
                                            beta: float,
                                            num_samples: int = 500) -> torch.Tensor:
        # * num_samples correspond to num of samples L in the paper
        # * note that for exact value for \hat \log \pi_\beta in the paper
        # we also need **an expection over L samples**
        mean, std = self.encode(state, action, goal)
        #print("mean.shape:",mean.shape)

        mean_enc = mean.repeat(num_samples, 1, 1).permute(1, 0, 2)  # [B x S x D]
        
        std_enc = std.repeat(num_samples, 1, 1).permute(1, 0, 2)  # [B x S x D]
        z = mean_enc + std_enc * torch.randn_like(std_enc)  # [B x S x D]

        state = state.repeat(num_samples, 1, 1).permute(1, 0, 2)  # [B x S x C]
        action = action.repeat(num_samples, 1, 1).permute(1, 0, 2)  # [B x S x C]
        goal_ = goal.repeat(num_samples, 1, 1).permute(1, 0, 2)  # [B x S x C]
        mean_dec = self.decode(state, action, z)

        std_dec = np.sqrt(beta / 4)
        # Find q(z|x)
        log_qzx = td.Normal(loc=mean_enc, scale=std_enc).log_prob(z)
        # Find p(z)
        mu_prior = torch.zeros_like(z).to(self.device)
        std_prior = torch.ones_like(z).to(self.device)
        log_pz = td.Normal(loc=mu_prior, scale=std_prior).log_prob(z)
        # Find p(x|z)
        std_dec = torch.ones_like(mean_dec).to(self.device) * std_dec
        #
        log_pxz = td.Normal(loc=mean_dec, scale=std_dec).log_prob(goal_)

        w = log_pxz.sum(-1) + log_pz.sum(-1) - log_qzx.sum(-1)
        ll = w.logsumexp(dim=-1) - np.log(num_samples)
        return ll

    def encode(self, state : torch.Tensor,
                     action: torch.Tensor,
                     goal  : torch.Tensor ) -> Tuple[torch.Tensor, torch.Tensor]:
        # goal: (batchsize, 1)
        z = self.encoder_shared(torch.cat([state, action, goal], -1))
        mean = self.mean(z)
        # Clamped for numerical stability
        log_std = self.log_std(z).clamp(-4, 15)
        std = torch.exp(log_std)
        return mean, std

    def decode(self, state: torch.Tensor,
                     action: torch.Tensor,
                     z: torch.Tensor = None,) -> torch.Tensor:
        # When sampling from the VAE, the latent vector is clipped to [-0.5, 0.5]
        # [state, action, z]-> -> prob (goal)
        if z is None:
            z = (torch.randn((state.shape[0], self.latent_dim)).to(self.device).clamp(-0.5, 0.5))
        mid_feature = torch.cat([action, state, z], -1)
        return self.decoder(mid_feature)
    
    def save_model(self, path: str):
        torch.save(self.state_dict(), path)

    def load_model(self, path: str):
        self.load_state_dict(torch.load(path))

def train(cfg, hydra_cfg):

    #set seed
    set_seed(cfg.seed)

    #set device
    device = torch.device(cfg.device)

    if cfg.save_snapshot:
        checkpoint_path = Path(hydra_cfg['runtime']['output_dir']) / Path('checkpoint')
        checkpoint_path.mkdir(exist_ok=True)
        log_path =  Path(hydra_cfg['runtime']['output_dir'])
        best_eval_returns = 0

    start_time = datetime.now().replace(microsecond=0)
    time_elapsed = start_time - start_time
    start_time_str = start_time.strftime("%y-%m-%d-%H-%M-%S")
    if cfg.dataset_name in ["pointmaze-umaze-v0", "pointmaze-medium-v0", "pointmaze-large-v0"]:
        cfg.vision = False
        if cfg.dataset_name in ["pointmaze-umaze-v0"]:
            cfg.env_name = 'PointMaze_UMaze-v3'
            cfg.nclusters = 20 if cfg.nclusters is None else cfg.nclusters
        elif cfg.dataset_name in ["pointmaze-medium-v0"]:
            cfg.env_name = 'PointMaze_Medium-v3'
            cfg.nclusters = 40 if cfg.nclusters is None else cfg.nclusters
        elif cfg.dataset_name == "pointmaze-large-v0":
            cfg.env_name = 'PointMaze_Large-v3'
            cfg.nclusters = 80 if cfg.nclusters is None else cfg.nclusters
        env = gym.make(cfg.env_name, continuing_task=False)
    
    elif "antmaze" in cfg.dataset_name:
        if "umaze" in cfg.dataset_name:
            cfg.env_name = 'AntMaze_UMaze-v4'
            cfg.nclusters = 20 if cfg.nclusters is None else cfg.nclusters
        elif "medium" in cfg.dataset_name:
            cfg.env_name = 'AntMaze_Medium-v4'
            cfg.nclusters = 40 if cfg.nclusters is None else cfg.nclusters
        elif "large" in cfg.dataset_name:
            cfg.env_name = 'AntMaze_Large-v4'
            cfg.nclusters = 80 if cfg.nclusters is None else cfg.nclusters
        env = AntmazeWrapper(gym.make(cfg.env_name, continuing_task=False))

    else:
        raise NotImplementedError
    
    env.action_space.seed(cfg.seed)
    env.observation_space.seed(cfg.seed)
   
    print(cfg.nclusters)
    
    train_dataset = MaxEpisodicDataset(env, cfg.dataset_name, cfg. datasize, cfg.augment_data, cfg.augment_prob, cfg.nclusters, cfg.vision)
        

    train_data_loader = DataLoader(
                            train_dataset,
                            batch_size=cfg.batch_size,
                            shuffle=True,
                            num_workers=cfg.num_workers
                        )
    train_data_iter = iter(train_data_loader)

    #create model
    model = DecisionRvs(cfg.env_name, env, goal_dim=train_dataset.goal_dim).to(device)

    optimizer = torch.optim.Adam(
                        model.parameters(),
                        lr=cfg.lr,
                    )
    
    v_model = DecisionV(cfg.env_name, env, goal_dim=train_dataset.goal_dim).to(device)


    try:
                #states, proprio, goals, actions = next(train_data_iter)
                state, goal, action = next(train_data_iter)

    except StopIteration:
        train_data_iter = iter(train_data_loader)
        #states, proprio, goals, actions = next(train_data_iter)
        state, goal, action = next(train_data_iter)

    state_dim = state.shape[-1]
    action_dim = action.shape[-1]
    goal_dim = goal.shape[-1]

    max_action = float(env.action_space.high[0])

    #VAE Train
    vae = VAE(
        state_dim, action_dim, goal_dim, 2 *goal_dim, max_action, cfg.vae_hidden_dim
    ).to(device)

    directory_path = os.path.join(cfg.vae_model_path, cfg.dataset_name)

    path = Path(directory_path)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
        print(f"Directory '{path}' created.")
    else:
        print(f"Directory '{path}' already exists.")

    model_path = os.path.join(path, 'vae_model.pth')
    if os.path.exists(model_path):
        print("Loading existing VAE model...")
        vae.load_model(model_path)
    else:
        print("Training VAE!!!!!!!!!!")
        vae_optimizer = torch.optim.Adam(vae.parameters(), lr=cfg.vae_lr)
        #cfg.vae_iterations
        for t in range(int(cfg.vae_iterations)):
            try:
                states, goal, actions = next(train_data_iter)
                #timesteps, states, goal, actions, traj_mask = next(train_data_iter)

            except StopIteration:
                train_data_iter = iter(train_data_loader)
                states, goal, actions = next(train_data_iter)
                #timesteps, states, goal, actions, traj_mask = next(train_data_iter)
            
            if cfg.vision:
                states = states.to(device).squeeze(-1).unsqueeze(2)                                  # B x T x pixel_dim
            else:
                states = states.to(device)       
            
            #proprio = proprio.to(device)                                # B x T x state_dim
            goal = goal.to(device).repeat(1, cfg.context_len, 1)      # B x T x goal_dim
            actions = actions.to(device)
            
            log_dict = {}
            # Variational Auto-Encoder Training
            recon, mean, std = vae(states, actions, goal)
            #print("recon.shape:",recon.shape)
            #print("goal.shape:",goal.shape)
            recon_loss = F.mse_loss(recon, goal)
            KL_loss = -0.5 * (1 + torch.log(std.pow(2)) - mean.pow(2) - std.pow(2)).mean()
            vae_loss = recon_loss + cfg.beta * KL_loss
    
            vae_optimizer.zero_grad()
            vae_loss.backward()
            vae_optimizer.step()
    
            log_dict["VAE/reconstruction_loss"] = recon_loss.item()
            log_dict["VAE/KL_loss"] = KL_loss.item()
            log_dict["VAE/vae_loss"] = vae_loss.item()
            log_dict["vae_iter"] = t
        print("train vae results:", log_dict)
        print("Saving VAE model...")
        vae.save_model(model_path)

    vae.eval()
    total_updates = 0
    success_rate_list = []
    for i_train_iter in range(cfg.max_train_iters):

        log_action_losses = []
        model.train()
        #cfg.num_updates_per_iter
        for i in range(cfg.num_updates_per_iter):   
            try:
                # states, proprio, goals, actions = next(train_data_iter)
                states, goal, actions = next(train_data_iter)

            except StopIteration:
                train_data_iter = iter(train_data_loader)
                # states, proprio, goals, actions = next(train_data_iter)
                states, goal, actions = next(train_data_iter)

            if cfg.vision:
                states = states.to(device).squeeze(-1).unsqueeze(1)
            else:
                states = states.to(device)                          # B x state_dim
            
            # proprio = proprio.to(device)
            actions = actions.to(device)                        # B
            goal = goal.to(device)                            # B x state_dim            

            #state_goals = torch.cat((states,goal),dim=-1).to(device)
            returns_to_go = vae.importance_sampling_estimator(states, actions, goal, cfg.beta).to(device).unsqueeze(
                dim=-1
            )                                          # B x T x 1

            #print("returns_to_go.shape:",returns_to_go.shape)
            returns_to_go_preds = v_model.forward(
                                states=states, 
                                #actions= actions,
                                goals=goal,
                            )

            # model forward ----------------------------------------------
            action_preds = model.forward(
                                states=states, 
                                # proprio=proprio,
                                goals=goal,
                                q= returns_to_go_preds,
                            )

            returns_to_go_target = torch.clone(returns_to_go)
            returns_to_go_preds = returns_to_go_preds

            # returns_to_go_loss -----------------------------------------
            norm = returns_to_go_target.abs().mean()
            u = (returns_to_go_target - returns_to_go_preds) / norm
            returns_to_go_loss = torch.mean(
                torch.abs(
                    cfg.tau - (u < 0).float()
                ) * u ** 2
            )
            
            action_loss = F.mse_loss(action_preds, actions, reduction='mean')

            loss = returns_to_go_loss + action_loss

            # optimization -----------------------------------------------
            optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(
                model.parameters(), 
                cfg.grad_norm
            )
            optimizer.step()


            log_action_losses.append(loss.detach().cpu().item())

        time = datetime.now().replace(microsecond=0) - start_time - time_elapsed
        time_elapsed = datetime.now().replace(microsecond=0) - start_time

        total_updates += cfg.num_updates_per_iter
        
        mean_action_loss = np.mean(log_action_losses)

        results = eval_env(cfg, model, v_model, device, render=cfg.render)

        log_str = ("=" * 60 + '\n' +
                "time elapsed: " + str(time_elapsed)  + '\n' +
                "num of updates: " + str(total_updates) + '\n' +
                "train action loss: " +  format(mean_action_loss, ".5f") #+ '\n' +
            )
        success_rate_list.append(results['eval/avg_reward'])
        print(log_str)
        print(results)
        print("eval_reward_list:", success_rate_list)

        if cfg.save_snapshot and results['eval/avg_reward'] >= best_eval_returns:
            print("=" * 60)
            print("saving best model!")
            print("=" * 60)
            best_eval_returns = results['eval/avg_reward']
            print("*******************************************************************************")
            print("total_updates:!!!!!!!!!!!!!!!!!",total_updates)
            print("*******************************************************************************")
            print("*******************************************************************************")
            print("best_eval_returns:!!!!!!!!!!!!!!!!!",best_eval_returns)
            print("*******************************************************************************")
    
    log_data = results
    log_filename = 'log.txt'
    if cfg.augment_data == True:
        new_log_filename = 'max_return_' +log_filename
    else:
        new_log_filename = 'no_augment_' +log_filename
    print("log_path:", log_path)
    
    log_file_path = os.path.join(log_path, new_log_filename)

    
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    
    with open(log_file_path, 'w') as file:
        for key, value in log_data.items():
            file.write(f'{key}: {value}\n')

        print(f'Data has been written to {log_file_path}')

    print("=" * 60)
    print("finished training!")
    print("=" * 60)
    end_time = datetime.now().replace(microsecond=0)
    time_elapsed = str(end_time - start_time)
    end_time_str = end_time.strftime("%y-%m-%d-%H-%M-%S")
    print("started training at: " + start_time_str)
    print("finished training at: " + end_time_str)
    print("total training time: " + time_elapsed)
    print("*******************************************************************************")
    print("total_updates:!!!!!!!!!!!!!!!!!", total_updates)
    print("*******************************************************************************")
    print("*******************************************************************************")
    print("eval_reward_list:!!!!!!!!!!!!!!!!!", success_rate_list)
    print("*******************************************************************************")
    print("*******************************************************************************")
    print("best_eval_returns:!!!!!!!!!!!!!!!!!",best_eval_returns)
    print("*******************************************************************************")
    print("=" * 60)
    
@hydra.main(config_path='cfgs', config_name='max_dmlp', version_base=None)
def main(cfg: DictConfig):
    hydra_cfg = hydra.core.hydra_config.HydraConfig.get()
    
    if cfg.wandb_log:
        if cfg.wandb_dir is None:
            cfg.wandb_dir = hydra_cfg['runtime']['output_dir']

        project_name = cfg.dataset_name
        wandb.init(project=project_name, entity=cfg.wandb_entity, config=dict(cfg), dir=cfg.wandb_dir, group=cfg.wandb_group_name)
        wandb.run.name = cfg.wandb_run_name
    
    train(cfg, hydra_cfg)
        
if __name__ == "__main__":
    main()